{-# LANGUAGE ScopedTypeVariables #-}
{-
  This module implements a naive converssion to CNF.
  It suffers from exponential growth, in case a DNF is
  passed as input.
  
  The Uniplate library makes the transformations very easy.
-}
module CNF ( 
    Literal(Plus, Minus), Clause, CNF,  -- types
    neg, litval, litAbs,                -- Operations over literals
    isCNF, isClause, isLiteral, hasTop, -- Predicates
    getCNF, newCNF, simplifyCNF,        -- operations over CNF's
    getSimpleCNF, showCNF
  ) where

import Props
import Data.Generics.Uniplate.Data
import Data.List(nub, sort, intersperse)
import qualified Data.Data as D

-- Some datatypes and instances are needed.
data Literal a 
  = Plus  a -- A simple variable
  | Minus a -- A negated variable
  deriving (Eq, Show)

instance (Ord a) => Ord (Literal a) where
  compare (Plus x) (Plus y) = compare x y
  compare (Plus _) (Minus _) = GT
  compare (Minus _) (Plus _) = LT
  compare (Minus x) (Minus y) = compare x y

-- We'll use the set notation for CNF, in haskell, list of lists.
type Clause a = [Literal a]
type CNF a = [Clause a]

showCNF :: CNF String -> String
showCNF = k . (intersperse ", ") . (map pclause)
    where
      pclause = parens . k . (intersperse ", ") . (map plit)
      plit (Plus s) = s
      plit (Minus s) = '~':s
      parens s = "(" ++ s ++ ")"
      k = foldr (++) ""

-- Some operations over literals might be needed, either internally or
-- by another module.
neg :: Literal a -> Literal a
neg (Plus x) = (Minus x)
neg (Minus x)= (Plus x)

litval :: Literal a -> a
litval (Plus x) = x
litval (Minus x) = x

litAbs :: Literal a -> Literal a
litAbs = Plus . litval

-- Some predicates over formulas and CNF's.
hasTop :: (Eq a) => Clause a -> Bool
hasTop [] = False
hasTop (l:ls) = ((neg l) `elem` ls) || (hasTop ls)

isLiteral :: PFormula a -> Bool
isLiteral (PNeg (PVar _)) = True
isLiteral (PVar _) = True
isLiteral _ = False
            
isClause :: PFormula a -> Bool
isClause (POr x y) = isClause x && isClause y
isClause x = isLiteral x
            
isCNF :: PFormula a -> Bool
isCNF (PAnd x y) = isCNF x && isCNF y
isCNF x@(POr _ _) = isClause x
isCNF x = isLiteral x

-- And now the transformations.
-- (minimalLogic f) returns Nothing if it can't convert the formula f,
-- returns a equivalent formula, converting one operator to a "better" one.
-- The converssions implemented are:
--    a -> b  == ~a || b
--    a iff b == (a -> b) && (b -> a)
--
minimalLogic :: PFormula a -> Maybe (PFormula a)
minimalLogic (PImp p q) = Just (POr (PNeg p) q)
minimalLogic (PIff p q) = Just (PAnd (PImp p q) (PImp q p))
minimalLogic x          = Nothing

-- De Morgan's and negation idempotency.
pushNegation :: PFormula a -> Maybe (PFormula a)
pushNegation (PNeg (PNeg p))   = Just p
pushNegation (PNeg (PAnd p q)) = Just (POr  (PNeg p) (PNeg q))
pushNegation (PNeg (POr  p q)) = Just (PAnd (PNeg p) (PNeg q))
pushNegation x                 = Nothing

-- distributive law
distribute :: PFormula a -> Maybe (PFormula a)
distribute (POr p (PAnd m n)) = Just (PAnd (POr p m) (POr p n))
distribute (POr (PAnd m n) q) = Just (PAnd (POr m q) (POr n q))
distribute x                  = Nothing

-- and now we use Uniplate's rewrite to recursively apply those transformation
-- until a "fixed point" is reached. That is, prepareCNF f = f.
-- Note.: the order of the transformations is important, since, for example,
--        pushNegation wont propagate a negation over a equivalence.
prepareCNF :: forall a. D.Data a => PFormula a -> PFormula a
prepareCNF = (rw (distribute :: PFormula a -> Maybe (PFormula a))) . 
             (rw (pushNegation :: PFormula a -> Maybe (PFormula a))) . 
             (rw (minimalLogic :: PFormula a -> Maybe (PFormula a)))
          where
            rw = rewrite
 
-- After prepareCNF, we have a CNF formula, with type (PFormula a).
-- we need to convert it to type (CNF a).           
getCNF :: (Ord a, Show a, Eq a) => PFormula a -> CNF a
getCNF = sepClauses
  where
    getClauses (PNeg (PVar s)) = [Minus s]
    getClauses (PVar s) = [Plus s]
    getClauses (POr p q) = (getClauses p) ++ (getClauses q)
    getClauses x         = error ("*** not part of a clause: " ++ show x)
    
    sepClauses (PAnd p q) = sepClauses p ++ sepClauses q
    sepClauses x          = [getClauses x]

-- Now we can parse a propositional formula and get either a parse error
-- or a CNF, this is done with (newCNF str)          
newCNF :: String -> Either String (CNF String)
newCNF s = either (Left . show) (Right . getSimpleCNF . prepareCNF) (parseFormula s)

-- The parsed CNF will have lot's of trivial clauses, duplicate literals, and
-- other redundancies. Those redundancies can be simplified.
simplifyCNF :: (Ord a, Show a, Eq a) => CNF a -> CNF a
simplifyCNF = nub . rmDuplicateClauses . rmTopClauses
  where
    rmDuplicateClauses = map (sort . nub)    
    rmTopClauses = filter (not . hasTop)

-- And for convenience, given a propositional formula, we may want to get
-- the full simplified CNF equivalent.              
getSimpleCNF :: (D.Data a, Ord a, Show a, Eq a) => PFormula a -> CNF a
getSimpleCNF = simplifyCNF . getCNF . prepareCNF
