module DIMACS (
  outDIMACS, inDIMACS -- IO functions to and for the DIMACS format.
  ) where
  
import CNF
import System.IO
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Language
import qualified Text.Parsec.Token as P

-- given a CNF, outputs a commented dimacs file.   
outDIMACS :: (Show a, Eq a) => CNF a -> String -> IO ()
outDIMACS cnf fname = let
    vars = cnfVars cnf
    idxv = makeIndex vars
    clauses = length cnf
  in do
    { h <- openFile fname WriteMode
    ; outDIMACSdecls h (length vars) clauses 
    ; outDIMACScomments h idxv
    ; outDIMACSclauses h idxv cnf
    ; hFlush h
    ; hClose h
    }
  where
    makeIndex v = zip v (take (length v) [1..])
    cnfVars = foldl clauseVars []
    clauseVars = foldl addLit
    addLit acu lit = if (litAbs lit) `elem` acu
                      then acu
                      else (litAbs lit):acu

-- Given a handle, a number of variables and a number of clauses, 
-- outputs the first DIMACS line, the declaration.    
outDIMACSdecls :: Handle -> Int -> Int -> IO ()
outDIMACSdecls h nv nc 
  = hPutStrLn h ("p cnf " ++ (show nv) ++ ' ':(show nc))
  >> hFlush h

-- Outputs the mapping used (Literal a -> Int) as comments.
outDIMACScomments :: (Show a) => Handle -> [(Literal a, Int)] -> IO ()
outDIMACScomments h m 
  = hPutStrLn h ("c map " ++ (show (length m)))
    >> foldr (\v r -> 
          hPutStrLn h ("c " ++ (show . litval . fst) v ++ " := " ++ (show . snd) v) >> r
       ) (return ()) m
    >> hFlush h

-- outputs the dimacs clauses
outDIMACSclauses :: (Show a, Eq a) => Handle -> [(Literal a, Int)] -> CNF a -> IO ()
outDIMACSclauses h m cnf = (sequence_ . (map (g h m))) cnf
  where
    g h1 m1 = (>> hPutStrLn h1 "0") . sequence_ . (map (f h1 m1))
    
    litSig (Plus _) = 1
    litSig (Minus _) = -1

    f h1 m1 lit = case lookup (litAbs lit) m1 of
                  (Just s) -> hPutStr h1 (show (s*(litSig lit)) ++ " ")
                  Nothing  -> error "*** DIMACS.outDIMACSclauses(lookup)"

-- input a DIMACS file, using parsec.                  
inDIMACS :: String -> IO (CNF String)
inDIMACS s = do
  { h <- openFile s ReadMode
  ; c <- hGetContents h
  -- ; hClose h
  ; case parse pDIMACS "" c of
      (Left err) -> putStrLn ("*** parse error: " ++ (show err)) >> return []
      (Right f)  -> return f
  }
  where
    lexer       = P.makeTokenParser emptyDef
    iden        = P.identifier lexer
    sym         = P.symbol lexer
    int         = P.integer lexer
    ws          = P.whiteSpace lexer
    
     
    pDIMACScomments :: Parser ()
    pDIMACScomments = try (sym "c")
                    >> many (satisfy (/= '\n'))
                    >> ws >> return ()
                       
    pDIMACSident :: Parser Int
    pDIMACSident = int >>= return . fromIntegral

      
    pDIMACS :: Parser (CNF String)
    pDIMACS = do
      { optional (many pDIMACScomments)
      ; pDIMACSdecls
      ; dict <- option [] pDIMACSdict
      ; many (pDIMACSclause dict)
      }
  
    pDIMACSdecls :: Parser ()
    pDIMACSdecls = sym "p" >> sym "cnf" >> pDIMACSident >> pDIMACSident >> return ()
    
    pDIMACSdict = pDIMACScomments >> ws >> many pDIMACSdictEntry
    pDIMACSdictEntry = sym "c" >> 
                       (iden <|> quotes iden) >>= \i ->
                       sym ":=" >>
                       pDIMACSident >>= \d ->
                       return (d, i)
                       
    quotes p = char '"' >> p >>= \i -> char '"' >> ws >> return i
                       
    pDIMACSclause :: [(Int, String)] -> Parser (Clause String)
    pDIMACSclause d = do
      { optional (many pDIMACScomments)
      ; i <- pDIMACSident
      ; if i == 0
          then return []
          else pDIMACSclause d >>= return . ((intToLit d i):)
      }
      
    intToLit d i = let 
      lit = maybe ("x" ++ show (abs i)) id (lookup (abs i) d) 
      in if i < 0
        then Minus lit
        else Plus lit
                                          
                                  
    
  

                                        
