{-
  This module implements the DPLL algorithm for SAT.
  It not only gives us a answer (SAT or UNSAT), but also a
  valuation that satisfies our formula.
  
  The input must be a CNF.
-}
module DPLL(dpll, dpllAssuming, Valuation) where

import Control.Monad.State
import Data.List(delete, union)
import CNF

-- Our valuation representation is simple, but not evident.
-- We'll use a list of literals as a valuation. So given a valuation val,
-- for every literal l, 
--    if l \in val then value(l) = TRUE
type Valuation a = Clause a
type DPLLst a = State (Valuation a) 

-- A unit is a literal that occurs alone in a clause,
-- this function searches the CNF for a unit.
findUnit :: CNF a -> Maybe (Literal a)
findUnit = foldr cond Nothing
  where
    cond h r = if length h == 1 
                then Just (head h)
                else r

-- |Given a CNF and a literal, this function will return a CNF
--  thet is the first one simplified, assuming that l holds.
simplifyClause :: (Eq a) => CNF a -> Literal a -> CNF a
simplifyClause s l = [ delete (neg l) c | c <- s, l `notElem` c ]
                
-- |Unit Propagation rule
unitPropagate :: (Eq a) => CNF a -> DPLLst a (CNF a)
unitPropagate s 
  = case findUnit s of
      Nothing -> return s
      Just l  -> if [] `elem` s then return s else do
        { v <- get
        ; put (v `union` [l])
        ; unitPropagate (simplifyClause s l)
        }
        
-- auxiliar version, using the state monad.       
dpllax :: (Eq a) => CNF a -> DPLLst a Bool
dpllax s = do
  { s' <- unitPropagate s
  ; if s' == []
      then return True
      else if [] `elem` s'
        then return False
        else let l = chooseLit s' in do
          { val <- get -- save the current state.
          ; put (val `union` [l])
          ; k <- dpllax (simplifyClause s' l)
          ; if k 
              then return k
              else do
                { put (val `union` [neg l])
                ; dpllax (simplifyClause s' (neg l))
                }           
          }
  }
  where
    chooseLit = head . head

-- calls the auxiliar version with one empty valuation.    
dpll :: (Eq a) => CNF a -> Maybe (Valuation a)
dpll s = case runState (dpllax s) [] of
            (True, l) -> Just l
            _         -> Nothing
            
-- calls the auxliar version, with a given valuation.
dpllAssuming :: (Eq a) => CNF a -> Valuation a -> Maybe (Valuation a)
dpllAssuming s val = dpll ((map (:[]) val) ++ s)
     
