{-# LANGUAGE ScopedTypeVariables #-}
{-
  Main module, implements user interaction.
-}
module Main where

-- pure imports:
-- import Control.Monad(foldM)
import Control.Monad.Trans
import Data.List(isPrefixOf)
import DPLL
import CNF

-- impure imports
import DIMACS
import System.IO
import System.Environment(getArgs)
import System.CPUTime

-- We'll use a monad transformer to keep track of our state
-- inside the IO monad.

-- First, we define our state's type.
data ST = ST { cnf :: CNF String
             , val :: Valuation String
             }
             deriving (Show)
             
-- then, we define a newtype (so we can create instances with it), similarly
-- as the state monad.
-- note that the definition of the StateT generalizes this case.
newtype MState m u = MState {run :: ST -> m (ST, u)}

-- MState is also a monad:
instance Monad m => Monad (MState m) where 
  return k    = MState $ \st -> return (st, k)
  m >>= n     = MState $ \st -> do (st1, k1) <- run m st
                                   (st2, k2) <- run (n k1) st1
                                   return (st2, k2)
-- and a monad transformer
instance MonadTrans MState where
-- the type of lift is (Monad m => m a -> MState m a), from this we can
-- already see what it should do. Given a blackbox m, we want to put it
-- inside our MState blackbox:
  lift k = MState $ \st -> do a <- k          -- open the box
                              return (st, a)  -- put it insde our box
                              
-- some utility functions now
getState :: Monad m => (ST -> k) -> MState m k
getState g = MState $ \st -> return (st, g st)

modState :: Monad m => (ST -> ST) -> MState m ()
modState g = MState $ \st -> return (g st, ())  

setCNF :: CNF String -> ST -> ST
setCNF f _ = ST f []

setVal :: Valuation String -> ST -> ST
setVal v (ST f _) = ST f v

-- and now we can code out interpreter. We'll do it
-- in a top-down fashion
main :: IO ()
main = getArgs
     >>= \args -> run (makeSeq (groupCmdLineArgs "-" args)) (ST [] [])
     >> putStrLn "bye!"
     where
       makeSeq [] = userint
       makeSeq ([]:ls) = makeSeq ls
       makeSeq ((h1:t1):ls) = runCommand h1 t1 >> makeSeq ls

-- Given a list l, for example: ["-dimacs", "file1", "-solve", "-pf"]
-- and a string, for example "-", groupCmdLineArgs "-" l will result in the
-- following list:
--    groupCmdLineArgs "-" l ==
--      [["dimacs", "file1"], ["solve"], ["pf"]]
groupCmdLineArgs :: String -> [String] -> [[String]]
groupCmdLineArgs p ss = 
  let 
    lp = length p
    f a s = if p `isPrefixOf` s
              then [drop lp s]:a
              else ((head a) ++ [s]):(tail a)
  in reverse (foldl f ([] :: [[String]]) ss)

-- sugaring
mPutStr, mPutStrLn :: String -> MState IO ()
mPutStr     = lift . putStr
mPutStrLn   = lift . putStrLn

mGetLine    :: MState IO String
mGetLine    = lift getLine

promptFull, promptEmpty :: MState IO ()
promptFull  = mPutStr "* >> "
promptEmpty = mPutStr "  >> "

promptError :: String -> MState IO ()
promptError = mPutStrLn . ("! error: " ++)



-- performs the interaction with the user
userint :: MState IO ()
userint = do
  { f <- getState cnf
  ; lift (hFlush stdout)
  ; prompt (f /= [])
  ; lift (hFlush stdout)
  ; i <- mGetLine
  ; if isCommand i
      then let (h:a) = words (tail i) in runCommand h a
      else setNewFormula i
  }
  where
    prompt True  = promptFull
    prompt False = promptEmpty
    
    isCommand = (== ':') . head

-- sets the a new formula    
setNewFormula :: String -> MState IO ()
setNewFormula s
  = case newCNF s of
      (Left err) -> promptError (show err) >> userint
      (Right f)  -> modState (setCNF f) >> userint
      
-- a more general lookup function
mlookup :: (Eq b) => b -> [([b], a)] -> Maybe a
mlookup _ [] = Nothing
mlookup s ((idx, v):t) | s `elem` idx = Just v
                       | otherwise    = mlookup s t
      
-- runs a given command
runCommand :: String -> [String] -> MState IO ()
runCommand [] _     = mPutStrLn "?" >> userint
runCommand "exit" _ = return ()
runCommand s a = maybe (err s) (mapply a) (mlookup s commands) >> userint
  where    
    err k = mPutStrLn ("unkown command: " ++ k) >> userint
    mapply v = either id ($ v)
    -- and now that we built the interpreter framework,
    -- part, just implement commands and register them.
    commands = [
      (["cf", "clearcnf"], Left $   modState (setCNF [])),
      (["cv", "clearval"], Left $   modState (setVal [])),
      (["a", "assume"],    Right $  cmdAssume),
      (["h", "help"],      Left $   cmdHelp),
      (["s", "solve"],     Left $   cmdSolve),
      (["pv", "printval"], Left $   cmdPrintVal),
      (["pf", "printcnf"], Left $   cmdPrintCNF),
      (["n", "next"],      Left $   cmdNextVal),
      (["dimacs"],         Right $  cmdDimacs)
      ]
      
      
-- command type declarations
cmdHelp, cmdSolve, cmdPrintVal, cmdPrintCNF, cmdNextVal :: MState IO ()
      
            
-- shows a help screen
cmdHelp = do
  { h <- lift (openFile "helpfile" ReadMode)
  ; s <- lift (hGetContents h)
  ; mPutStrLn "helpfile:"
  ; mPutStrLn s
  ; lift (hClose h)
  }

-- solves the current formula
cmdSolve = do
  { t1 <- lift getCPUTime
  ; f <- getState cnf
  ; v <- getState val
  ; r <- (maybe (promptError "UNSAT" >> return False) 
                ((>> return True) . modState . setVal)
          ) (dpllAssuming f v)
  ; t2 <- lift getCPUTime
  ; mPutStrLn (show ((fromIntegral (t2 - t1))*(1E-6 :: Double)) ++ " microseconds (10^-6 s) elapsed.")
  ; if r 
      then cmdPrintVal
      else return ()
  }
         
-- etc ...
cmdPrintVal = getState val
            >>= \v -> mPutStrLn ('\n':(foldr ((++).plit) "" v))
  where
    plit (Plus a)  = ('\t':a) ++ "\t\t" ++ "True\n" 
    plit (Minus a) = ('\t':a) ++ "\t\t" ++ "False\n"
    
cmdPrintCNF = getState cnf >>= mPutStrLn . showCNF
    
cmdNextVal = do
  { v <- getState val
  ; modState $ setVal (map neg v)
  ; cmdSolve
  }       

cmdAssume :: [String] -> MState IO ()
cmdAssume = modState . setVal . (map p)
  where
    p :: String -> Literal String
    p s | head s `elem` "-~" = Minus (tail s)
        | otherwise          = Plus  s
        
cmdDimacs :: [String] -> MState IO ()
cmdDimacs [s] = lift (inDIMACS s)
              >>= modState . setCNF
cmdDimacs _   = mPutStrLn ":dimacs <filename>"

