{-# LANGUAGE DeriveDataTypeable #-}
{-
  This module implements a simple parser for propositional formulas.
  Very straighforward given some Parsec knowledge.
  
  The Data and Typeable instances are needed later on, on the CNF module.
-}
module Props where

-- easy imports
import Control.Applicative hiding (many, (<|>), optional)

-- cabal install parsec uniplate
import Text.ParserCombinators.Parsec.Expr
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Language
import qualified Text.Parsec.Token as P
import qualified Data.Data as D

data PFormula a
  = PVar a
  | PNeg (PFormula a)
  | PAnd (PFormula a) (PFormula a)
  | POr  (PFormula a) (PFormula a)
  | PImp (PFormula a) (PFormula a)
  | PIff (PFormula a) (PFormula a)
  deriving (Eq, Show, D.Data, D.Typeable)


logicStyle :: LanguageDef st
logicStyle = emptyDef
                { commentStart   = "#-"
                , commentEnd     = "-#"
                , commentLine    = "#"
                , nestedComments = True
                , identStart     = letter
                , identLetter	 = alphaNum <|> oneOf "_'"
                , reservedOpNames= ["*", "&&", "+", "||", "-", "~",
                                    "->", "<->"]
                , reservedNames  = []
                , caseSensitive  = True
                }

lexer       = P.makeTokenParser logicStyle  
identifier  = P.identifier lexer
reservedOp  = P.reservedOp lexer
parens      = P.parens lexer

parseFormula :: String -> Either String (PFormula String)
parseFormula s = case parse pFormula "" s of
                    (Left err) -> (Left (show err))
                    (Right f)  -> (Right f)

pOpAnd :: Parser ()
pOpAnd = reservedOp "*" <|> reservedOp "&&"

pOpOr :: Parser ()
pOpOr = reservedOp "+" <|> reservedOp "||"

pOpImp :: Parser ()
pOpImp = reservedOp "->"

pOpIff :: Parser ()
pOpIff = reservedOp "<->"

pOpNeg :: Parser ()
pOpNeg = (reservedOp "-" >> notFollowedBy (char '>'))
        <|> reservedOp "~"
        
pAtomicExp :: Parser (PFormula String)
pAtomicExp = (PVar <$> identifier) <|> (parens pFormula)

pFormula :: Parser (PFormula String)
pFormula = buildExpressionParser table pAtomicExp
  where
    table =
      [[Prefix (pOpNeg >> return PNeg)],
       [Infix (pOpAnd >> return PAnd) AssocLeft],
       [Infix (pOpOr  >> return POr) AssocLeft],
       [Infix (pOpImp >> return PImp) AssocRight],
       [Infix (pOpIff >> return PIff) AssocNone]
      ]




